
//		SafeChars.cpp - plugin for reading/writing text containing 'dangerous' characters


#define UNICODE				1
#define _UNICODE			1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <winbase.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR			1
#define VER_MINOR			10
#define VER_BUILD			1
#define VER_PATCH			0

#define PLUGIN_NAME			L"SafeChars"

#ifdef _WIN64
#define PLATFORM			L"x64"
#else
#define PLATFORM			L"x86"
#endif


const wchar_t safequote		= 0xff02;
const wchar_t safepct		= 0xff05;
const wchar_t safeamp		= 0xff06;
const wchar_t safeapos		= 0xff07;
const wchar_t safeopat		= 0xff08;
const wchar_t safecpat		= 0xff09;
const wchar_t safelss		= 0xff1c;
const wchar_t safegtr		= 0xff1e;
const wchar_t safeobrk		= 0xff3b;
const wchar_t safecbrk		= 0xff3d;
const wchar_t safecaret		= 0xff3e;
const wchar_t safegrave		= 0xff40;
const wchar_t safevbar		= 0xff5c;

const wchar_t CR			= 0x000d;
const wchar_t LF			= 0x000a;
const wchar_t QUOTE			= 0x0022;
const wchar_t APOS			= 0x0027;


#define MAX_ARG_LEN			16384
#define SHORT_BUF_LEN		1024
#define MAX_ARGS			1024

#define MAX_FILENAME_LEN	16384

#define ERRHANDLE			GetStdHandle( STD_ERROR_HANDLE )



const wchar_t PluginName[] = L"SafeChars";
wchar_t PluginFeatures[] = L"@countsafe,@countsafex,@countunsafe,@countunsafex,@safecharsinfo,@safeclip,@safeclipw,\
@safeenv,@safeexp,@safeline,@saferead,@safewrite,@tagattr,@unsafe,@unsafeesc,@uquotes,\
_amp,_bq,_caret,_closebrk,_closepat,_gt,_lt,_openbrk,_openpat,_pct,_quote,_vbar,\
safearray,safecharshelp,safeecho,safeechos,unsafe";


const wchar_t RegKeyName[]   = L"SOFTWARE\\JPPlugins\\SafeChars";
const wchar_t RegValueName[] = L"Unsafe";


#define UNSAFE_MIN		32
#define UNSAFE_MAX		127
#define UNSAFE_RANGE	96

#define REMAP_MIN		0xff00
#define REMAP_MAX		0xff5f
#define REMAP_DIFF		0xfee0

#define COUNT_SAFECHARS		0
#define COUNT_UNSAFECHARS	1
#define COUNT_OTHERCHARS	2



BOOL Unsafe[UNSAFE_RANGE];
BOOL Unsafe2[UNSAFE_RANGE];


wchar_t svSetDos[32] = L"";
wchar_t fxSetDos[32] = L"";

HINSTANCE MyDLLHandle			= NULL;



const wchar_t SafeArrayHelpText[] = L"Replace dangerous characters or quote elements in an array.\n\
\n\
SAFEARRAY /C:n /R:n /Q /QE arrayname\r\n\
\n\
\t/C:  Column (2-D arrays only)\n\
\t/R:  Row    (2-D arrays only)\n\
\t/Q   Quote elements instead of replacing characters\n\
\t/QE   ...  including Empty elements\n\
\n\
The arrayname is required.  The array must already exist; it\n\
must have one or two dimensions.  If /C: is not specified,\n\
all columns will be affected; if /R: is not specified, all\n\
rows will be affected.\n";


const wchar_t SafeCharsHelpHelpText[] = L"Open the SafeChars plugin help file.\n\
\n\
SAFECHARSHELP /C /F /S /S:text /V topic\r\n\
\n\
\t/C\tselect the Contents tab\n\
\t/F\tselect the Favorites tab\n\
\t/S\tselect the Search tab\n\
\t/S:\tselect the Search tab and search for text\n\
\t/V\tshow detailed plugin version info\n\
\n\
Opens this plugin\u2019s help file to the specified topic.\n\
If no topic is given, opens the Overview page by default.  \n";


const wchar_t SafeEchoHelpText[] = L"Display remapped text to standard output.\n\
\n\
SAFEECHO text\r\n\
\n\
Displays text to standard output, restoring any  \n\
remapped \u2018safe\u2019 characters to their original,\n\
possibly dangerous values.\n";


const wchar_t SafeEchosHelpText[] = L"Display remapped text to standard output.\n\
\n\
SAFEECHOS text\r\n\
\n\
Displays text to standard output, restoring any\n\
remapped \u2018safe\u2019 characters to their original,\n\
possibly dangerous values.  A carriage return and  \n\
line feed will not be appended.\n";


const wchar_t UnsafeHelpText[] = L"Enable, disable, save, restore, or list dangerous characters.\n\
\n\
UNSAFE /D:chars /E:chars /R /S /Z\r\n\
\n\
\t/D:  Disable (treat characters as safe)\n\
\t/E:  Enable  (treat characters as dangerous)\n\
\t/R   Restore list from registry\n\
\t/S   Save list to registry\n\
\t/Z   Reset to plugin defaults\n\
\n\
chars may be:  a single character, a value 32 - 127, or a  \n\
hex value 0x20 - 0x7f; a range, two characters separated\n\
by a minus sign; or a list of characters/ranges separated\n\
by commas or semicolons.\n\
\n\
/S saves the current list, making it the default when the\n\
plugin is reloaded in the future.  /R restores from the\n\
registry; /Z resets to the original defaults.  /R and /Z\n\
are mutually exclusive.\n\
\n\
If no options are given, UNSAFE will display the current\n\
list of dangerous characters to standard output.\n";



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( GetStdHandle(STD_ERROR_HANDLE), L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( GetStdHandle(STD_ERROR_HANDLE), shortbuf );

	return 0;
}



#include "ParseArgs.cpp"



int SaveToRegistry()
{
	HKEY KeyHandle = NULL;
	BYTE Buffer[12];
	
	memset( Buffer, 0x00, sizeof( Buffer ) );
	for ( int i = 0; i < UNSAFE_RANGE; i++ ) {
		int j = i / 8;
		int k = 1 << ( i % 8 );
		if ( Unsafe[i] )
			Buffer[j] |= k;
	}

	int rv = 0;

	if ( rv = RegCreateKeyEx( HKEY_CURRENT_USER, RegKeyName, 0, NULL, 0, KEY_ALL_ACCESS, NULL, &KeyHandle, NULL ) )
		return rv;

	rv = RegSetValueEx( KeyHandle, RegValueName, 0, REG_BINARY, Buffer, 12 );

	RegCloseKey( KeyHandle );
	return rv;
}


int RestoreFromRegistry()
{
	HKEY KeyHandle = NULL;
	BYTE Buffer[12];
	DWORD ValueType = 0, ValueSize = 12;

	int rv = 0;

	if ( rv = RegOpenKeyEx( HKEY_CURRENT_USER, RegKeyName, 0, KEY_READ, &KeyHandle ) )
		return rv;

	rv = RegQueryValueEx( KeyHandle, RegValueName, NULL, &ValueType, Buffer, &ValueSize );
	RegCloseKey( KeyHandle );

	if ( rv )
		return rv;
	else if ( ( ValueType != REG_BINARY ) || ( ValueSize != 12 ) )
		return ERROR_INVALID_DATA;

	for ( int i = 0; i < UNSAFE_RANGE; i++ ) {
		int j = i / 8;
		int k = 1 << ( i % 8 );
		Unsafe[i] = ( ( Buffer[j] & k ) != 0 );
	}

	return 0;
}


int DefaultUnsafeChars()
{
	memset( Unsafe, 0x00, sizeof( Unsafe ) );

	Unsafe[ '"' - 0x20 ] = TRUE;
	Unsafe[ '%' - 0x20 ] = TRUE;
	Unsafe[ '&' - 0x20 ] = TRUE;
	Unsafe[ '(' - 0x20 ] = TRUE;
	Unsafe[ ')' - 0x20 ] = TRUE;
	Unsafe[ '<' - 0x20 ] = TRUE;
	Unsafe[ '>' - 0x20 ] = TRUE;
	Unsafe[ '[' - 0x20 ] = TRUE;
	Unsafe[ ']' - 0x20 ] = TRUE;
	Unsafe[ '^' - 0x20 ] = TRUE;
	Unsafe[ '`' - 0x20 ] = TRUE;
	Unsafe[ '|' - 0x20 ] = TRUE;

	return 0;
}


int Stash()
{
	for ( int i = 0; i < UNSAFE_RANGE; i++ )
		Unsafe2[i] = Unsafe[i];

	return 0;
}


int Unstash()
{
	for ( int i = 0; i < UNSAFE_RANGE; i++ )
		Unsafe[i] = Unsafe2[i];

	return 0;
}


BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
			MyDLLHandle = (HINSTANCE) hModule;
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}


int DisableOneFeature( LPTSTR FeaturesList, LPTSTR FeatureToDisable )
{
	if ( !FeaturesList || !FeatureToDisable )
		return -1;

	if ( ( FeaturesList[0] == '\0' ) || ( FeatureToDisable[0] == '\0' ) )
		return -1;


	size_t i = 0, j = 0, base = 0;
	wchar_t ch = '\0', ch2 = '\0';
	BOOL done = FALSE, match = FALSE;

	do {
		// Compare word in 'FeatureToDisable' against word at 'base' :

		i = 0;
		match = TRUE;

		do {
			ch = towupper( FeaturesList[base + i] );
			ch2 = ch;
			if ( ( ch2 == ',' ) || ( ch2 == ';' ) )
				ch2 = '\0';

			match &= ( ch2 == towupper( FeatureToDisable[i] ) );

			i++;
		} while ( ch2 != '\0' );

		if ( match ) {
			// disable the feature:

			if ( ch == '\0' )
				FeaturesList[base] = ch;
			else {
				j = 0;

				do {
					ch = FeaturesList[base + i + j];
					FeaturesList[base + j] = ch;
					j++;
				} while ( ch != '\0' );
			}
		}
		else
			base += i;

	} while ( ch != '\0' );

	i = wcslen( FeaturesList );
	if ( i > 0 )
		if ( ( FeaturesList[i - 1] == ',' ) || ( FeaturesList[i - 1] == ';' ) )
			FeaturesList[i - 1] = '\0';

	return 0;
}


BOOL ArraysAvailable( )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"";
	int ShellMajor = -1, TccLe = 0;
	wchar_t *sep1 = NULL;

	wcscpy_s( shortbuf, SHORT_BUF_LEN, L"%_4ver" );
	ExpandVariables( shortbuf, 1 );
	ShellMajor = wcstol( shortbuf, &sep1, 10 );

	wcscpy_s( shortbuf, SHORT_BUF_LEN, L"%_cmdproc" );
	ExpandVariables( shortbuf, 1 );
	if ( !_wcsicmp( shortbuf, L"TCCLE" ) )
		TccLe = 1;

	if ( ( ShellMajor < 10 ) || ( ShellMajor > 255 ) )
		return FALSE;

	if ( TccLe )
		return FALSE;

	return TRUE;
}


#include "NewHelp.cpp"


DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	if ( !ArraysAvailable() )
		DisableOneFeature( PluginFeatures, L"safearray" );

	static PLUGININFO piInfo;

	piInfo.pszDll = (LPTSTR)L"SafeChars";
	piInfo.pszAuthor = (LPTSTR)L"Charles Dye";
	piInfo.pszEmail = (LPTSTR)L"cdye@unm.edu";
	piInfo.pszWWW = (LPTSTR)L"http://charlesdye.net/plugins/safechars.html";
	piInfo.pszDescription = (LPTSTR)L"Read and write text containing special characters";
	piInfo.pszFunctions = (LPTSTR) PluginFeatures;
	piInfo.nMajor = VER_MAJOR;
	piInfo.nMinor = VER_MINOR;
	piInfo.nBuild = VER_BUILD;

	return &piInfo;
}


DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%_4ver";
	ExpandVariables( shortbuf, 1 );

	if ( shortbuf[0] == '7' )
		if ( iswpunct( shortbuf[1] ) ) {
			DisplayErrorHeader();
			Qprintf(GetStdHandle(STD_ERROR_HANDLE), L"4NT / Take Command v8.0 or higher required!\r\n");
			return -1;
		}

	wcscpy_s( shortbuf, SHORT_BUF_LEN, L"%[_transient]%[_pipe]%[noloadmsg]" );
	ExpandVariables( shortbuf, 1 );

	DefaultUnsafeChars();

	if ( RestoreFromRegistry() == ERROR_FILE_NOT_FOUND )
		SaveToRegistry();

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s plugin v%d.%d.%d loaded.\n", PluginName, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}


DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	CloseHelpWindow();
	return 0;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned long int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned decimal

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar = sep1;

		return 0;
	}

	return -1;
}


int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( !FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) )
		Sprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}


//	Calls an internal TCC command.

int CallInternal( LPTSTR Cmd, LPTSTR Args )
{
	if ( !Cmd )
		return -1;
	else if ( Cmd[0] == '\0' )
		return -1;


	wchar_t CmdBuf[MAX_ARG_LEN]		= L"";
	wchar_t ExpVar[SHORT_BUF_LEN]	= L"%_expansion";
	BOOL AddStar					= TRUE;


	ExpandVariables( ExpVar, 1 );
	if ( wcschr( ExpVar, '1' ) )
		AddStar = FALSE;


	if ( ( Cmd[0] != '*' ) && AddStar )
		wcscpy_s( CmdBuf, MAX_ARG_LEN, L"*" );

	wcscat_s( CmdBuf, MAX_ARG_LEN, Cmd );

	if ( Args ) {
		if ( Args[0] ) {
			if ( !iswspace( Args[0] ) )
				wcscat_s( CmdBuf, MAX_ARG_LEN, L" " );

			wcscat_s( CmdBuf, MAX_ARG_LEN, Args );
		}
	}

	return Command( CmdBuf, 0 );
}


int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



int ConsoleWidth( )
{
	if ( !QueryIsConsole( GetStdHandle( STD_OUTPUT_HANDLE ) ) )
		return 0;

	HANDLE ConsoleHandle = CreateFile( L"CONOUT$", ( GENERIC_READ | GENERIC_WRITE ), ( FILE_SHARE_READ | FILE_SHARE_WRITE ), 0, OPEN_EXISTING, NULL, NULL );
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	int Width = 0, rv = 0;

	if ( ConsoleHandle != INVALID_HANDLE_VALUE ) {
		rv = GetConsoleScreenBufferInfo( ConsoleHandle, &ConsoleInfo );
		CloseHandle( ConsoleHandle );

		if ( rv )
			Width = ConsoleInfo.dwSize.X;
	}

	return Width;
}


BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}


BOOL IsListSep( wchar_t ch )
{
	return ( ( ch == ',' ) || ( ch == ';' ) );
}


BOOL IsUnsafe( wchar_t ch )
{
	if ( ( ch < UNSAFE_MIN ) || ( ch > UNSAFE_MAX ) )
		return FALSE;

	return Unsafe[ch - UNSAFE_MIN];
}


BOOL IsFraction( wchar_t ch )
{
	if ( ( ch >= 0xbc ) && ( ch <= 0xbe ) )
		return TRUE;
	else if ( ( ch >= 0x2150 ) && ( ch <= 0x215e ) )
		return TRUE;
	else
		return FALSE;
}


BOOL IsVowel2( wchar_t ch )
{
	ch = towlower( ch );
	if ( ( ch == 'a' ) || ( ch == 'e' ) || ( ch == 'i' ) || ( ch == 'o' ) || ( ch == 'u' ) )
		return TRUE;

	if ( ( ch == 0x0100 ) || ( ch == 0x0101 ) || ( ch == 0x0112 ) || ( ch == 0x0113 ) || ( ch == 0x012a ) || ( ch == 0x012b ) || ( ch == 0x014c ) || ( ch == 0x014d ) || ( ch == 0x016a ) || ( ch == 0x016b ) )
		return TRUE;		//	vowels with macrons -- used in hawai'ian

	return FALSE;
}


BOOL IsEOL2( wchar_t ch )
{
	return ( ( ch == CR ) || ( ch == LF ) || ( ch == '\0' ) );
}


int CopySafeChars( LPTSTR source, LPTSTR target )
{
	if ( !source || !target )
		return -1;


	int i = 0, k = 0;
	wchar_t ch = '\0';

	do {
		ch = source[i];

		if ( IsUnsafe( ch ) ) {
			ch += REMAP_DIFF;
			k++;
		}

		if ( i >= ( MAX_ARG_LEN - 1 ) )
			ch = '\0';

		target[i] = ch;
		i++;

	} while ( ch != '\0' );

	return k;
}


int RestoreUnsafeChars( LPTSTR source, LPTSTR target )
{
	if ( !source || !target )
		return -1;


	int i = 0, k = 0;
	wchar_t ch = '\0';

	do {
		ch = source[i];

		if ( ( ch >= REMAP_MIN ) && ( ch <= REMAP_MAX ) ) {
			ch -= REMAP_DIFF;
			k++;
		}

		if ( i >= ( MAX_ARG_LEN - 1 ) )
			ch = '\0';

		target[i] = ch;
		i++;

	} while ( ch != '\0' );

	return k;
}


int PrepSetDos()
{
	wcscpy_s(svSetDos, 32, L"%_expansion");

	ExpandVariables(svSetDos, 1);
	if ( svSetDos[0] == '\0' )
		wcscpy_s(svSetDos, 32, L"0");

	CallInternal( L"setdos", L" /x-1478+3" );			//	Setdos_Cmd( L" /x-1478+3");

	return 0;
}


int ResetSetDos()
{
	wcscpy_s(fxSetDos, 32, L"/X");
	wcscat_s(fxSetDos, 32, svSetDos);

	CallInternal( L"setdos", fxSetDos );				//	Setdos_Cmd( fxSetDos );

	return 0;
}


int CountCharacterTypes( LPTSTR inString, int Type, BOOL Expand )
{
	if ( !inString )
		return -1;


	int SafeChars = 0, UnsafeChars = 0, OtherChars = 0;
	wchar_t ch = 0;

	if ( Expand ) {

		LPTSTR VarName = inString;
		wchar_t TempString[MAX_ARG_LEN] = L"";
		int rv = 0;

		if ( VarName[0] == '%' )
			VarName++;

		if ( VarName[0] == '\0' ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing variable or function name\r\n" );
			return -1;
		}

		wcscpy_s( TempString, MAX_ARG_LEN, L"%" );
		wcscat_s( TempString, MAX_ARG_LEN, VarName );
		TempString[MAX_ARG_LEN - 1] = 0x00;

		PrepSetDos();
		rv = ExpandVariables( TempString, 1 );
		ResetSetDos();

		if ( rv < 0 )
			return rv;

		if ( TempString[0] == '\0' )
			return 0;

		wcscpy_s( inString, MAX_ARG_LEN, TempString );
	}

	for ( int i = 0; inString[i]; i++ ) {
		ch = inString[i];

		if ( ( ch >= REMAP_MIN ) && ( ch <= REMAP_MAX ) )
			SafeChars++;
		else if ( IsUnsafe( ch ) )
			UnsafeChars++;
		else
			OtherChars++;
	}
	
	if ( Type == COUNT_SAFECHARS )
		return SafeChars;
	else if ( Type == COUNT_UNSAFECHARS )
		return UnsafeChars;
	else if ( Type == COUNT_OTHERCHARS )
		return OtherChars;
	else
		return -2;
}


DLLExports int WINAPI f_countsafe( LPTSTR inString )
{
	int rv =  CountCharacterTypes( inString, COUNT_SAFECHARS, FALSE );

	if ( rv < 0 )
		return rv;
	
	wsprintf( inString, L"%d", rv );
	return 0;
}


DLLExports int WINAPI f_countsafex( LPTSTR inString )
{
	int rv = CountCharacterTypes( inString, COUNT_SAFECHARS, TRUE );

	if ( rv < 0 )
		return rv;
	
	wsprintf( inString, L"%d", rv );
	return 0;
}


DLLExports int WINAPI f_countunsafe( LPTSTR inString )
{
	int rv = CountCharacterTypes( inString, COUNT_UNSAFECHARS, FALSE );

	if ( rv < 0 )
		return rv;
	
	wsprintf( inString, L"%d", rv );
	return 0;
}


DLLExports int WINAPI f_countunsafex( LPTSTR inString )
{
	int rv = CountCharacterTypes( inString, COUNT_UNSAFECHARS, TRUE );

	if ( rv < 0 )
		return rv;
	
	wsprintf( inString, L"%d", rv );
	return 0;
}


DLLExports int WINAPI f_safecharsinfo( LPTSTR inString )
{
	if ( !inString )
		return -1;


	unsigned int Select = 0;
	wchar_t *sep1 = NULL;

	if ( ParseInt( inString, &Select, &sep1 ) ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad value \"%s\"\r\n", inString );
		return -1;
	}

	switch ( Select ) {

		case 0 :
			wsprintf( inString, L"%u.%u.%u", VER_MAJOR, VER_MINOR,VER_BUILD );
			break;

		case 1 :
			inString[0] = 0x00;
			unsigned int Word;

			for ( int i = UNSAFE_RANGE - 1; i >= 0; i-- ) {
				if ( ( i % 16 ) == 15 )
					Word = 0;

				Word = ( Word << 1 ) | ( Unsafe[i] ? 1 : 0 );

				if ( ( i % 16 ) == 0 )
					wsprintf( inString + (int) wcslen( inString ), L"%04X", Word );
			}

			break;

		case 2 :
			inString[0] = 0x00;

			if ( ( *sep1 == ',' ) || ( *sep1 == ';' ) )
				CopySafeChars( sep1 + 1, inString );

			break;

		case 3 :
			inString[0] = 0x00;

			if ( ( *sep1 == ',' ) || ( *sep1 == ';' ) )
				RestoreUnsafeChars( sep1 + 1, inString );

			break;

		default :
			wsprintf( inString, L"?%u", Select );
			break;
	}

	return 0;
}


DLLExports int WINAPI f_safeclip( LPTSTR inString )
{
	if ( !inString )
		return 1;

	wchar_t TempString[MAX_ARG_LEN] = L"";

	int rv = 0;

	wcscpy_s(TempString, MAX_ARG_LEN, L"%@clip[");
	wcscat_s(TempString, MAX_ARG_LEN, inString);
	wcscat_s(TempString, MAX_ARG_LEN, L"]");
	TempString[MAX_ARG_LEN - 1] = '\0';

	PrepSetDos();
	rv = ExpandVariables(TempString, 1);
	ResetSetDos();

	if ( rv < 0 )
		return rv;

	CopySafeChars(TempString, inString);
	return 0;
}


DLLExports int WINAPI f_safeclipw( LPTSTR inString )
{
	if ( !inString )
		return 1;

	wchar_t ch = '*';
	int rv = 0;

	RestoreUnsafeChars( inString, inString );

	rv = CopyTextToClipboard( inString , (int) wcslen(inString) );

	inString[0] = rv ? '1' : '0';
	inString[1] = '\0';
	return 0;
}


DLLExports int WINAPI f_safeenv( LPTSTR inString )
{
	if ( !inString )
		return 1;

	LPTSTR VarName = inString;
	wchar_t TempString[MAX_ARG_LEN] = L"";

	DWORD rv = 0;
	if ( VarName[0] == '%' )
		VarName++;

	if ( VarName[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"No variable name\r\n" );
		return -87;
	}

	rv = GetEnvironmentVariable( VarName, TempString, MAX_ARG_LEN );
	TempString[MAX_ARG_LEN - 1] = '\0';

	if ( !rv ) {
		inString[0] = '\0';
		return 0;
	}

	CopySafeChars( TempString, inString );
	return 0;
}


DLLExports int WINAPI f_safeexp( LPTSTR inString )
{
	if ( !inString )
		return 1;


	LPTSTR VarName = inString;
	wchar_t TempString[MAX_ARG_LEN] = L"";

	int rv = 0;
	if ( VarName[0] == '%' )
		VarName++;

	if ( VarName[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing variable or function name\r\n" );
		return -87;
	}

	if ( !_wcsnicmp( VarName, L"@safeexp[", 9 ) ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Don\u2019t make my brain explode!\r\n" );
		return -87;
	}

	wcscpy_s( TempString, MAX_ARG_LEN, L"%" );
	wcscat_s( TempString, MAX_ARG_LEN, VarName );
	TempString[MAX_ARG_LEN - 1] = '\0';

	PrepSetDos();
	rv = ExpandVariables( TempString, 0 );
	ResetSetDos();

	if ( rv < 0 )
		return rv;

	CopySafeChars( TempString, inString );
	return 0;
}


DLLExports int WINAPI f_unsafe( LPTSTR inString )
{
	if ( !inString )
		return 1;

	RestoreUnsafeChars( inString, inString );
	return 0;
}


int EscapeUnsafeChars( LPTSTR String )
{
	if ( !String )
		return -1;


	wchar_t Temp[MAX_ARG_LEN] = L"";
	unsigned int j = 0;
	int k = 0;
	wchar_t ch = 0x00;


	for ( unsigned int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( IsUnsafe( ch ) ) {
			if ( j < ( MAX_ARG_LEN - 1 ) )
				Temp[j++] = '%';

			if ( ( ch != '%' ) && ( j < ( MAX_ARG_LEN - 1 ) ) )
				Temp[j++] = '=';

			if ( ch == '`' )
				ch = 'k';
			else if ( ch == '"' )
				ch = 'q';

			k++;
		}

		if ( j < ( MAX_ARG_LEN - 1 ) )
			Temp[j++] = ch;
	}

	wcscpy_s( String, MAX_ARG_LEN, Temp );
	return k;
}


DLLExports int WINAPI f_unsafeesc( LPTSTR String )
{
	if ( !String )
		return 1;


	RestoreUnsafeChars( String, String );
	EscapeUnsafeChars( String );
	return 0;
}


DLLExports int WINAPI f_safeline( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t TempString[MAX_ARG_LEN] = L"";

	int rv = 0;

	wcscpy_s(TempString, MAX_ARG_LEN, L"%@line[");
	wcscat_s(TempString, MAX_ARG_LEN, inString);
	wcscat_s(TempString, MAX_ARG_LEN, L"]");
	TempString[MAX_ARG_LEN - 1] = '\0';

	PrepSetDos();
	rv = ExpandVariables(TempString, 1);
	ResetSetDos();

	if ( rv < 0 )
		return rv;

	CopySafeChars(TempString, inString);
	return 0;
}


DLLExports int WINAPI f_saferead( LPTSTR inString )
{
	if ( !inString )
		return 1;

	wchar_t TempString[MAX_ARG_LEN] = L"";

	int rv = 0;

	wcscpy_s(TempString, MAX_ARG_LEN, L"%@fileread[");
	wcscat_s(TempString, MAX_ARG_LEN, inString);
	wcscat_s(TempString, MAX_ARG_LEN, L"]");
	TempString[MAX_ARG_LEN - 1] = '\0';

	PrepSetDos();
	rv = ExpandVariables(TempString, 1);
	ResetSetDos();

	if ( rv < 0 )
		return rv;

	CopySafeChars(TempString, inString);
	return 0;
}


unsigned int GetTccCodePage()
{
	wchar_t shortbuf[SHORT_BUF_LEN]		= L"%_codepage";
	unsigned int cp						= 0;


	ExpandVariables( shortbuf, 1 );

	if ( iswdigit( shortbuf[0] ) ) {
		_set_errno( 0 );
		cp = wcstoul( shortbuf, NULL, 10 );
		if ( errno )
			cp = 0;
	}

	if ( !cp )
		cp = GetOEMCP();

	return cp;
}


DLLExports int WINAPI f_safewrite( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t TempString[MAX_ARG_LEN] = L"";
	int i = 0, j = 0;
	wchar_t ch = '\0';
	wchar_t *sep1 = &ch;
	HANDLE FileHandle = NULL;
	DWORD BytesWritten = 0;
	int rv = 0, chars = 0;


	//  Copy the first parameter (the file handle) to TempString:

	do {
		ch = inString[i];
		i++;

		if ( iswspace( ch ) && !j )
			continue;

		if ( ( ch == ',' ) || ( j >= ( MAX_ARG_LEN - 1 ) ) )
			ch = '\0';

		TempString[j] = ch;
		j++;

	} while ( ch != '\0' );

	if ( TempString[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf(GetStdHandle(STD_ERROR_HANDLE), L"No handle\r\n");
		return -87;
	}
	else if ( ( inString[i - 1] != ',' ) && ( inString[i - 1] != '\0' ) ) {
		DisplayErrorHeader();
		Qprintf(GetStdHandle(STD_ERROR_HANDLE), L"No text argument\r\n");
		return -87;
	}

	// Parse TempString and save the file handle:

	if ( ( TempString[0] == '0' ) && ( tolower(TempString[1]) == 'x' ) && iswxdigit(TempString[2]) )
		FileHandle = (HANDLE) wcstol(TempString, &sep1, 16);
	else if ( iswdigit(TempString[0]) )
		FileHandle = (HANDLE) wcstol(TempString, &sep1, 10);
	else if ( !wcscmp( TempString, L"-" ) )
		FileHandle = GetStdHandle( STD_OUTPUT_HANDLE );

	if ( !FileHandle || *sep1 ) {
		DisplayErrorHeader();
		Qprintf(GetStdHandle(STD_ERROR_HANDLE), L"Bad handle \"%s\"\r\n", TempString);
		return -87;
	}

	// Copy the remainder of the input string (the text) to
	// TempString, mapping fullwidth characters back to
	// ASCII as we go:

	if ( inString[i - 1] == '\0' )
		i--;

	j = 0;
	do {
		ch = inString[i];
		i++;

		if ( ( ch >= REMAP_MIN ) && ( ch <= REMAP_MAX ) )
			ch -= REMAP_DIFF;

		if ( j >= ( MAX_ARG_LEN - 1 ) )
			ch = '\0';

		TempString[j] = ch;
		j++;

	} while ( ch != '\0' );

	if ( j > ( MAX_ARG_LEN - 2 ) ) {
		DisplayErrorHeader();
		Qprintf(GetStdHandle(STD_ERROR_HANDLE), L"String too long\r\n");
		return -87;
	}

	TempString[j-1] = CR;
	TempString[j]   = LF;
	TempString[j+1] = '\0';
	chars = j + 1;

	// Now write TempString to the file:

	if ( QueryUnicodeOutput() == 1 ) {
		// Do Unicode stuff ....
		BY_HANDLE_FILE_INFORMATION fi;
		BYTE bom[2] = { 0xff, 0xfe };
		rv = GetFileInformationByHandle(FileHandle, &fi);
		if ( rv )
			if ( ( fi.nFileSizeLow | fi.nFileSizeHigh ) == 0 )
				if ( !WriteFile( FileHandle, bom, 2, &BytesWritten, NULL ) ) {
					wcscpy_s( inString, MAX_ARG_LEN, L"-1" );
					return 0;
				}

		rv = WriteFile( FileHandle, TempString, chars * 2, &BytesWritten, NULL );
	}
	else {
		// Do 8-bit stuff ....
		char UTF8OutputBuffer[MAX_ARG_LEN * 4] = "";

		WideCharToMultiByte( GetTccCodePage(), 0, TempString, -1, UTF8OutputBuffer, MAX_ARG_LEN, NULL, NULL);
		rv = WriteFile(FileHandle, UTF8OutputBuffer, chars, &BytesWritten, NULL);
	}

	if ( rv )
		Sprintf(inString, L"%d", chars);
	else
		wcscpy_s(inString, MAX_ARG_LEN, L"-1");

	return 0;
}


BOOL IsWordBeginningWithApostrophe( LPTSTR inString )
{
	if ( !inString )
		return FALSE;
	else if ( inString[0] == '\0' )
		return FALSE;

	wchar_t Temp[SHORT_BUF_LEN] = L"";
	wchar_t ch = 0x00;
	int i = 0;

	const int NumWords = 21;
	const wchar_t Word[NumWords][12] = { L"'cos", L"'cus", L"'em", L"'mid", L"'mongst",
		L"'n", L"'n'", L"'neath", L"'tain't", L"'til", L"'tis", L"'tisn't", L"'twas", L"'twasn't",
		L"'tween", L"'twere", L"'twern't", L"'twixt", L"'twould", L"'twouldn't", L"'un" };

	do {
		ch = towlower( inString[i] );
		if ( ( ch == 0xff07 ) || ( ch == 0x2018 ) || ( ch == 0x2019 ) )
			ch = 0x27;

		if ( !iswalpha( ch ) && ( ch != 0x27 ) )
			ch = '\0';

		if ( i < ( SHORT_BUF_LEN - 1 ) )
			Temp[i++] = ch;

		Temp[i] = '\0';

	} while ( ch && ( i < ( SHORT_BUF_LEN - 1 ) ) );

	for ( i = 0; i < NumWords; i++ )
		if ( !wcscmp( Temp, Word[i] ) )
			return TRUE;

	return FALSE;
}


int GetEnvironmentChar( LPTSTR Name, wchar_t *outChar )
{
	if ( !Name )
		return -1;
	else if ( Name[0] == '\0' )
		return 0;

	wchar_t Temp[SHORT_BUF_LEN] = L"";
	unsigned int NumericChar = 0;
	int i = 0;
	wchar_t *sep1 = NULL;

	if ( GetEnvironmentVariable( Name, Temp, SHORT_BUF_LEN ) < 1 )
		return 0;
	else if ( Temp[0] == '\0' )
		return 0;

	if ( ( Temp[0] > 31 ) && ( Temp[1] == '\0' ) )
		if ( !iswdigit( Temp[0] ) ) {
			*outChar = Temp[0];
			return 0;
		}

	while ( iswspace( Temp[i] ) )
		i++;

	if ( !iswdigit( Temp[i] ) )
		return -1;

	if ( ParseInt( Temp + i, &NumericChar, &sep1 ) )
		return -1;

	while ( iswspace( *sep1 ) )
		sep1++;

	if ( *sep1 )
		return -1;

	if ( ( NumericChar < 32 ) || ( NumericChar > 0xffff ) )
		return -1;

	*outChar = NumericChar;
	return 0;
}


DLLExports int WINAPI f_uquotes( LPTSTR inString )
{
	if ( !inString )
		return 1;
	else if ( inString[0] == '\0' )
		return 0;

	wchar_t ch = 0x20, lastch = 0x00, nextch = 0x00;
	wchar_t OpenQuote = 0x201c,		CloseQuote = 0x201d;
	wchar_t OpenSQuote = 0x2018,	CloseSQuote = 0x2019;
	wchar_t Prime = APOS,			DoublePrime = QUOTE;
	wchar_t Apostrophe = 0x2019,	Okina = 0x2018;
	wchar_t EmDash = 0x2014;

	wchar_t Temp[MAX_ARG_LEN] = L"";
	int j = 0;

	GetEnvironmentChar( L"OPENQUOTE",   &OpenQuote );
	GetEnvironmentChar( L"CLOSEQUOTE",  &CloseQuote );
	GetEnvironmentChar( L"OPENSQUOTE",  &OpenSQuote );
	GetEnvironmentChar( L"CLOSESQUOTE", &CloseSQuote );
	GetEnvironmentChar( L"APOSTROPHE",  &Apostrophe );
	GetEnvironmentChar( L"PRIME",		&Prime );
	GetEnvironmentChar( L"DOUBLEPRIME",	&DoublePrime );
	GetEnvironmentChar( L"'OKINA",		&Okina );


	for ( int i = 0; inString[i]; i++  ) {
		lastch = ch;
		ch = inString[i];
		nextch = inString[i + 1];

		if ( ( ch == QUOTE ) || ( ch == safequote ) ) {
			if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswspace( nextch ) || IsEOL2( nextch ) ) )
				ch = DoublePrime;
			else if ( iswgraph( lastch ) && ( lastch != OpenSQuote ) )
				ch = CloseQuote;
			else {
				ch = OpenQuote;

				if ( iswspace( lastch ) )
					if ( IsEOL2( nextch ) )
						ch = CloseQuote;
			}
		}
		else if ( ( ch == APOS ) || ( ch == safeapos ) ) {
			if ( iswgraph( lastch ) && ( lastch != OpenQuote ) ) {
				ch = CloseSQuote;

				if ( iswalnum( lastch ) && iswalpha( nextch ) )
					ch = Apostrophe;
			}
			else {
				ch = OpenSQuote;

				if ( iswspace( lastch ) )
					if ( IsEOL2( nextch ) )
						ch = CloseSQuote;
			}

			if ( !iswalpha( lastch ) && iswalpha( nextch ) ) {
				if ( IsWordBeginningWithApostrophe( inString + i ) )
					ch = Apostrophe;
			}
			else if ( IsVowel2( lastch ) && IsVowel2( nextch ) )
				ch = Okina;

			if ( !iswalnum( lastch ) && iswdigit( nextch ) )
				if ( iswdigit( inString[i + 2] ) )
					if ( !iswalnum( inString[i + 3] ) )
						ch = Apostrophe;

			if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswdigit( nextch ) || IsFraction( nextch ) || iswspace( nextch ) || IsEOL2( nextch ) ) )
				ch = Prime;
		}
		else if ( ( ch == '-' ) || ( ch == 0xff0d ) ) {
			if ( ( ( lastch != '-' ) && ( lastch != 0xff0d ) ) && ( nextch == ch ) )
				if ( ( inString[i + 2] != '-' ) && ( inString[i + 2] != 0xff0d ) ) {
					ch = EmDash;
					i++;
				}
		}

		if ( j < ( MAX_ARG_LEN - 1 ) ) {
			Temp[j++] = ch;
			Temp[j] = 0x00;
		}
	}

	wcscpy_s( inString, MAX_ARG_LEN, Temp );
	return 0;
}



DLLExports int WINAPI f_tagattr( LPTSTR String )			//		parse through an html tag and return the specified attribute:
{
	if ( !String )
		return 1;


	wchar_t AttribName[SHORT_BUF_LEN]			= L"";
	wchar_t AttribString[MAX_ARG_LEN]			= L"";

	wchar_t *ch									= String;
	unsigned int j								= 0;
	BOOL BadChar								= FALSE;
	wchar_t quote								= '\0';
	int equals									= -1;


	while ( iswspace( *ch ) )			//	skip over any leading whitespace
		ch++;

	if ( !iswalnum( *ch ) ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing attribute name\r\n" );
		return -1;
	}

	while ( *ch ) {						//	copy the attribute name
		if ( IsListSep( *ch ) || iswspace( *ch ) || ( *ch == '\0' ) )
			break;

		if ( !iswalnum( *ch ) )			//	anything other than letters and digits is bad
			BadChar = TRUE;

		if ( j < ( SHORT_BUF_LEN - 1 ) ) {
			AttribName[j++] = towlower( *ch );
			AttribName[j] = '\0';
		}

		ch++;
	}

	if ( !j ) {							//	no attribute name is an error
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing attribute name\r\n" );
		return -1;
	}
	else if ( BadChar ) {				//	invalid characters in attribute name is also an error
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad characters in attribute name:  \"%s\"\r\n", AttribName );
		return -1;
	}


	while ( iswspace( *ch ) )			//	skip over any trailing whitespace
		ch++;

	if ( IsListSep( *ch ) )				//	skip over the comma
		ch++;

	while ( iswspace( *ch ) )			//	skip over any whitespace after the comma
		ch++;


	String[0] = '\0';					//	empty the function's output buffer

	if ( ( *ch != '<' ) && ( *ch != safelss ) )			//	the next character ought to be a less-than sign
		return 0;										//	otherwise, the html tag is missing, so return an empty string


	//	wprintf( L"   * Attribute name = '%s'   Tag = '%s'\n", AttribName, ch );


	ch++;								//	skip over the less-than sign
	while ( iswalpha( *ch ) )			//	and the tag name
		ch++;

	while ( iswspace( *ch ) )			//	skip over any whitespace following the tag name
		ch++;



	j					= 0;			//	empty the attribute-string buffer
	AttribString[j]		= '\0';
	equals				= -1;			//	equals sign not found yet


	do {								//	parse through the html tag ....

		if ( !quote && ( ( *ch == APOS ) || ( *ch == safeapos ) ) )							//	open single-quote
			quote = APOS;
		else if ( !quote && ( ( *ch == QUOTE ) || ( *ch == safequote ) ) )					//	open double-quote
			quote = QUOTE;
		else if ( ( quote == APOS ) && ( ( *ch == APOS ) || ( *ch == safeapos ) ) )			//	close single-quote
			quote = '\0';
		else if ( ( quote == QUOTE ) && ( ( *ch == QUOTE ) || ( *ch == safequote ) ) )		//	close double-quote
			quote = '\0';
		else if ( !quote && ( iswspace( *ch ) || ( *ch == '>' ) || ( *ch == safegtr ) ) ) {		//	unquoted space or greater-than sign ends the attribute string

			if ( j ) {
				//	wprintf( L"   * Found attribute:  %s\n", AttribString );

				if ( equals > 0 ) {
					if ( !_wcsnicmp( AttribString, AttribName, equals ) ) {					//	if the attribute name matches,
						//	wprintf( L"   * Correct!\n" );
						wcscpy_s( String, MAX_ARG_LEN, AttribString + equals + 1 );			//	copy the value to the function output buffer
						return 0;															//	and return success
					}
				}
			}

			j = 0;							//	otherwise, empty the attribute string buffer
			AttribString[j] = '\0';
			equals = -1;					//	and keep looking
		}
		else if ( j < ( MAX_ARG_LEN - 1 ) ) {
			if ( !quote && ( equals < 0 ) && ( *ch == '=' ) )		//	found an equals sign?
				equals = j;											//	remember its position

			AttribString[j++] = *ch;								//	add the character to the attribute-string buffer
			AttribString[j] = '\0';
		}


		if ( *ch == '\0' )				//	null character ends the search
			return 0;

		ch++;							//	otherwise, keep going
	}
	while ( TRUE );						//	forever!
}



int ReturnChar( LPTSTR outString, wchar_t ch)
{
	if ( !outString )
		return -1;


	outString[0] = ch;
	outString[1] = '\0';
	return 0;
}


DLLExports int WINAPI _amp( LPTSTR outString )
{
	if ( !outString )
		return 1;

	ReturnChar(outString, safeamp);
	return 0;
}


DLLExports int WINAPI _bq( LPTSTR outString )
{
	if ( !outString )
		return 1;

	ReturnChar(outString, safegrave);
	return 0;
}


DLLExports int WINAPI _caret( LPTSTR outString )
{
	if ( !outString )
		return 1;

	ReturnChar(outString, safecaret);
	return 0;
}


DLLExports int WINAPI _closebrk( LPTSTR outString )
{
	if ( !outString )
		return 1;

	ReturnChar( outString, safecbrk );
	return 0;
}


DLLExports int WINAPI _closepat( LPTSTR outString )
{
	if ( !outString )
		return 1;

	ReturnChar( outString, safecpat );
	return 0;
}


DLLExports int WINAPI _gt( LPTSTR outString )
{
	if ( !outString )
		return 1;

	ReturnChar(outString, safegtr);
	return 0;
}


DLLExports int WINAPI _lt( LPTSTR outString )
{
	if ( !outString )
		return 1;

	ReturnChar(outString, safelss);
	return 0;
}


DLLExports int WINAPI _openbrk( LPTSTR outString )
{
	if ( !outString )
		return 1;

	ReturnChar( outString, safeobrk );
	return 0;
}
 
DLLExports int WINAPI _openpat( LPTSTR outString )
{
	if ( !outString )
		return 1;

	ReturnChar( outString, safeopat );
	return 0;
}


DLLExports int WINAPI _pct( LPTSTR outString )
{
	if ( !outString )
		return 1;

	ReturnChar(outString, safepct);
	return 0;
}


DLLExports int WINAPI _quote( LPTSTR outString )
{
	if ( !outString )
		return 1;

	ReturnChar(outString, safequote);
	return 0;
}


DLLExports int WINAPI _vbar( LPTSTR outString )
{
	if ( !outString )
		return 1;

	ReturnChar(outString, safevbar);
	return 0;
}


int SafeEcho( LPTSTR inString, BOOL AddCrLf )
{
	if ( !inString )
		return 1;

	wchar_t Temp[MAX_ARG_LEN] = L"";
	int j = 0;

	while ( iswspace( inString[j] ) )
		j++;

	//  check for syntax request:

	if ( !wcscmp( inString + j, L"/?" ) )
		return ShowCmdHelp( AddCrLf ? SafeEchoHelpText : SafeEchosHelpText, FALSE );
	else if ( !wcscmp( inString + j, L"/??" ) )
		return ShowCmdHelp( AddCrLf ? SafeEchoHelpText : SafeEchosHelpText, TRUE );

	j = 0;
	if ( iswspace( inString[j] ) )
		j++;

	RestoreUnsafeChars( inString + j, Temp );
	if ( AddCrLf )
		wcscat_s( Temp, L"\r\n" );

	QPuts( Temp );
	return 0;
}


DLLExports int WINAPI safeecho( LPTSTR inString )
{
	return SafeEcho( inString, TRUE );
}


DLLExports int WINAPI safeechos( LPTSTR inString )
{
	return SafeEcho( inString, FALSE );
}



int ArrayInfo( LPTSTR ArrayName, int Which )
{
	if ( !ArrayName )
		return -1;
	else if ( ArrayName[0] == '\0' )
		return -1;

	if ( ( Which < 0 ) || ( Which > 5 ) )
		return -1;

	wchar_t shortbuf[SHORT_BUF_LEN] = L"";
	int rv = 0;

	swprintf( shortbuf, SHORT_BUF_LEN, L"%%@arrayinfo[%s,%d]", ArrayName, Which );
	ExpandVariables( shortbuf, 1 );

	rv = wcstol( shortbuf, NULL, 10 );
	return rv;
}


DLLExports int WINAPI safearray( LPTSTR inString )
{
	if ( !inString )
		return 1;


	LPTSTR ArrayName = NULL;
	unsigned int Column = 0, Row = 0;
	BOOL UserColumn = FALSE, UserRow = FALSE;
	int QuoteElements = 0;
	BOOL RestoreUnsafe = FALSE;

	wchar_t Element[MAX_ARG_LEN]    = L"";
	wchar_t SafeString[MAX_ARG_LEN] = L"";
	wchar_t CmdString[MAX_ARG_LEN]  = L"";
	wchar_t QuoteStr[2] = { 0x22, 0x00 };
	int rv = 0;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) ) {
				Unstash();
				return ShowCmdHelp( SafeArrayHelpText, FALSE );
			}
			else if ( !wcscmp( arg[i], L"/??" ) ) {
				Unstash();
				return ShowCmdHelp( SafeArrayHelpText, TRUE );
			}
			else if ( !_wcsnicmp( arg[i], L"/C", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Column, NULL ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad column value \"%s\"\r\n", arg[i] );
					return 1;
				}

				UserColumn = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/R", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Row, NULL ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad row value \"%s\"\r\n", arg[i] );
					return 1;
				}

				UserRow = TRUE;
			}
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				QuoteElements = 1;
			else if ( !_wcsicmp( arg[i], L"/QE" ) )
				QuoteElements = 2;
			else if ( !_wcsicmp( arg[i], L"/U" ) )
				RestoreUnsafe = TRUE;
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unknown option \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( !ArrayName )
				ArrayName = arg[i];
			else {
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}

	if ( !ArrayName ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing array name\r\n" );
		return 1;
	}

	_wcsupr_s( ArrayName, MAX_ARG_LEN );

	BOOL GoodName = TRUE;
	for ( int i = 0; ArrayName[i]; i++ ) {
		if ( !iswalnum( ArrayName[i] ) )
			if ( ( ArrayName[i] != '_' ) && ( ArrayName[i] != '$' ) )
				GoodName = FALSE;
	}
	if ( iswdigit( ArrayName[0] ) )
		GoodName = FALSE;

	if ( !GoodName ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad array name \"%s\"\r\n", ArrayName );
		return 1;
	}

	int Dimensions = ArrayInfo( ArrayName, 0 );

	if ( Dimensions < 1 ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Not an array \"%s\"\r\n", ArrayName );
		return 1;
	}
	else if ( Dimensions > 2 ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Too many dimensions \"%s\"\r\n", ArrayName );
		return 1;
	}

	int Rows = ArrayInfo( ArrayName, 1 );

	int Columns = 1;
	if ( Dimensions > 1 )
		Columns = ArrayInfo( ArrayName, 2 );
	else
		UserColumn = FALSE;


	int RowStart = 0, RowEnd = Rows - 1;
	int ColStart = 0, ColEnd = Columns - 1;

	if ( UserRow )
		RowStart = RowEnd = Row;

	if ( UserColumn )
		ColStart = ColEnd = Column;


	PrepSetDos();

	for ( int r = RowStart; r <= RowEnd; r++ )
		for ( int c = ColStart; c <= ColEnd; c++ ) {
			if ( Dimensions == 1 ) {
				wsprintf( Element, L"%%%s[%d]", ArrayName, r );
				wsprintf( CmdString, L"%s[%d]=", ArrayName, r );
				ExpandVariables( Element, 1 );
			}
			else {
				wsprintf( Element, L"%%%s[%d,%d]", ArrayName, r, c );
				wsprintf( CmdString, L"%s[%d,%d]=", ArrayName, r, c );
				ExpandVariables( Element, 1 );
			}

			if ( QuoteElements ) {
				wcscat_s( CmdString, MAX_ARG_LEN, QuoteStr );
				wcscat_s( CmdString, MAX_ARG_LEN, Element );
				wcscat_s( CmdString, MAX_ARG_LEN, QuoteStr );

				if ( Element[0] || ( QuoteElements > 1 ) )
					CallInternal( L"set", CmdString );					//	Set_Cmd( CmdString );
			}
			else if ( RestoreUnsafe ) {
				if ( RestoreUnsafeChars( Element, SafeString ) ) {
					CallInternal( L"setdos", L" /x-1378" );				//	Setdos_Cmd( L" /x-1378");
					wcscat_s( CmdString, MAX_ARG_LEN, SafeString );
					CallInternal( L"setdos", L" /x-1478+3" );			//	Setdos_Cmd( L" /x-1478+3");
					CallInternal( L"set", CmdString );					//	Set_Cmd( CmdString );
				}
			}
			else {
				if ( CopySafeChars( Element, SafeString ) ) {
					wcscat_s( CmdString, MAX_ARG_LEN, SafeString );
					CallInternal( L"set", CmdString );					//	Set_Cmd( CmdString );
				}
			}
		}

	ResetSetDos();
	return 0;
}


int ParseChar( wchar_t **ch )
{
	unsigned int rv = 0;
	wchar_t *sep1 = NULL;

	if ( !iswdigit( **ch ) || ( !iswdigit( *( *ch + 1 ) ) && ( towupper( *( *ch + 1 ) ) != 'X' ) ) ) {
		sep1 = ( *ch + 1 );
		if ( *sep1 && ( *sep1 != '-' ) && !IsListSep( *sep1 ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error" );
			return -1;
		}

		rv = **ch;

		if ( ( rv < UNSAFE_MIN ) || ( rv > UNSAFE_MAX ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Character must be in range %u - %u", UNSAFE_MIN, UNSAFE_MAX );
			return -1;
		}

		*ch = sep1;
		return rv;
	}

	if ( ParseInt( *ch, &rv, &sep1 ) ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error" );
		return -1;
	}

	if ( ( rv < UNSAFE_MIN ) || ( rv > UNSAFE_MAX ) ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Character must be in range %u - %u", UNSAFE_MIN, UNSAFE_MAX );
		return -1;
	}

	*ch = sep1;
	return rv;
}


int EnableChars( LPTSTR inString, BOOL Enable )
{
	if ( !inString )
		return -1;
	else if ( inString[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error" );
		return -1;
	}

	wchar_t *ch = inString;
	int Char1 = 0, Char2 = 0;

	while ( *ch ) {

		Char1 = ParseChar( &ch );
		if ( Char1 < 0 )
			return -1;

		Char2 = Char1;

		if ( *ch == '-' ) {
			ch++;
			if ( *ch == '\0' ) {
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error" );
				return -1;
			}

			Char2 = ParseChar( &ch );
			if ( Char2 < 0 )
				return -1;
		}

		if ( Char1 > Char2 ) {
			int tmp = Char1;
			Char1 = Char2;
			Char2 = tmp;
		}

		for ( int i = Char1; i <= Char2; i++ )
			Unsafe[i - UNSAFE_MIN] = Enable;

		if ( *ch ) {
			if ( !IsListSep( *ch ) ) {
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error" );
				return -1;
			}
			else
				ch++;
		}
	}

	return 0;
}


int ListChars()
{
	int Width = ConsoleWidth();
	if ( ( Width < 40 ) || ( Width > 256 ) )
		Width = 80;

	wchar_t Output[SHORT_BUF_LEN] = L"";
	int n = 0, m = 0;
	BOOL FoundAny = FALSE;
	wchar_t ch = 0x00;

	QPuts( L"\nUnsafe characters:\n" );

	for ( int i = 0; i < UNSAFE_RANGE; i++ ) {
		ch = i + UNSAFE_MIN;

		if ( Unsafe[i] ) {
			Sprintf( Output, L"'%c' (%u, 0x%02x)  ", ch, ch, ch );

			m = (int) wcslen( Output );

			if ( ( n + m ) >= Width ) {
				CrLf();
				n = 0;
			}

			n += m;
			QPuts( Output );
			FoundAny = TRUE;
		}
	}

	if ( FoundAny )
		CrLf();
	else
		QPuts( L"- none -\r\n" );

	return 0;
}


DLLExports int WINAPI unsafe( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_QUOTES_KLUDGE );


	BOOL FoundAny = FALSE;
	BOOL Save = FALSE, Restore = FALSE, Reset = FALSE;

	wchar_t ErrorText[SHORT_BUF_LEN] = L"";
	int err = 0;


	for ( int i = 0; i < numargs; i++ )
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) ) {
				Unstash();
				return ShowCmdHelp( UnsafeHelpText, FALSE );
			}
			else if ( !wcscmp( arg[i], L"/??" ) ) {
				Unstash();
				return ShowCmdHelp( UnsafeHelpText, TRUE );
			}
			else if ( !_wcsicmp( arg[i], L"/R" ) )
				Restore = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Z" ) )
				Reset = TRUE;
		}


	if ( Restore && Reset ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can\u2019t combine /R and /Z\r\n" );
		return 1;
	}

	if ( Restore ) {
		int err = RestoreFromRegistry();
		if ( err ) {
			GetErrorText( err, ErrorText );
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Error restoring data from registry:\r\n" );
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"%s\r\n", ErrorText );
			return 1;
		}
		else
			QPuts( L"Restored.\r\n" );
	}
	else if ( Reset ) {
		DefaultUnsafeChars();
		QPuts( L"Reset to plugin defaults.\r\n" );
	}

	Stash();

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !_wcsnicmp( arg[i], L"/D", 2 ) || !_wcsnicmp( arg[i], L"/E", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;
				
				if ( EnableChars( arg[i] + j, towupper( arg[i][1] ) == 'E' ) ) {
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L" \"%s\"\r\n", arg[i] );
					Unstash();
					return 1;
				}

				FoundAny = TRUE;
			}
			else if ( !_wcsicmp( arg[i], L"/S" ) )
				Save = TRUE;
			else if ( !_wcsicmp( arg[i], L"/R" ) )
				Restore = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Z" ) )
				Reset = TRUE;
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unknown option \"%s\"\r\n", arg[i] );
				Unstash();
				return 1;
			}
			else {
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error \"%s\"\r\n", arg[i] );
				Unstash();
				return 1;
			}

		}
	}


	if ( Save ) {
		int err = SaveToRegistry();
		if ( err ) {
			GetErrorText( err, ErrorText );
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Error saving data to registry:\r\n" );
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"%s\r\n", ErrorText );
			return 1;
		}
		else
			QPuts( L"Saved.\r\n" );
	}

	if ( Save || Restore || Reset || FoundAny )
		return 0;

	ListChars();
	return 0;
}



DLLExports int WINAPI safecharshelp( LPTSTR inString )
{
	return PluginHelpCmd( inString, SafeCharsHelpHelpText );
}



DLLExports int WINAPI Usage(LPTSTR pszName, LPTSTR pszUsage)
{
	if ( !pszName || !pszUsage )
		return 0;

	if ( !_wcsicmp( pszName, L"safearray" ) ) {
		wcscpy_s( pszUsage, SHORT_BUF_LEN, UsageText( SafeArrayHelpText ) );
		return 1;
	}
	else if ( !_wcsicmp( pszName, L"safecharshelp" ) ) {
		wcscpy_s( pszUsage, SHORT_BUF_LEN, UsageText( SafeCharsHelpHelpText ) );
		return 1;
	}
	else if ( !_wcsicmp( pszName, L"safeecho" ) ) {
		wcscpy_s( pszUsage, SHORT_BUF_LEN, UsageText( SafeEchoHelpText ) );
		return 1;
	}
	else if ( !_wcsicmp( pszName, L"safeechos" ) ) {
		wcscpy_s( pszUsage, SHORT_BUF_LEN, UsageText( SafeEchosHelpText ) );
		return 1;
	}
	else if ( !_wcsicmp( pszName, L"unsafe" ) ) {
		wcscpy_s( pszUsage, SHORT_BUF_LEN, UsageText( UnsafeHelpText ) );
		return 1;
	}

	return 0;
}

